# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import unicode_literals

import numpy as np

from mlencoders.base_encoder import BaseEncoder
from mlencoders.base_encoder import NAN_CATEGORY


class WeightOfEvidenceEncoder(BaseEncoder):
    """
    Weight of Evidence Encoder for categorical features and binary outcome.

    For a categorical variable X, with possible values X_i, i=1..n, and target binary variable Y,
    it is defined as

        WOE_i = ln[ P(X=X_i | Y=1) / P(X=X_i | Y=0) ]

    """

    def __init__(self, cols=None, handle_unseen='impute', min_samples=1):
        """Instantiation

        :param [str] cols: list of columns to encode, or None (then all dataset columns will be encoded at fitting time)
        :param str handle_unseen:
            'impute' - default value, impute a -1 category
            'error'  - raise an error if a category unseen at fitting time is found
            'ignore' - skip unseen categories
        :param int min_samples: minimum samples to compute WOE of category, must be >= 1.

        :return: None
        """
        super(WeightOfEvidenceEncoder, self).__init__(cols, handle_unseen, min_samples, 0)

    def fit(self, X, y):
        """Encode given columns of X according to y.

        :param pandas.DataFrame X: DataFrame of features, shape (n_samples, n_features). Must contain columns to encode.
        :param pandas.Series y: pandas Series of target values, shape (n_samples,).

        :return: None
        """
        self._before_fit_check(X, y)
        for col in self.cols:
            # Share of positive (resp. negative) labels for each category P(X=X_i | Y=1) (resp. P(X=X_i | Y=0))
            mapping = y.groupby(X[col].fillna(NAN_CATEGORY)).agg(['sum', 'count']).rename({'sum': 'pos'}, axis=1)
            mapping['neg'] = mapping['count'] - mapping['pos']
            mapping[['pos', 'neg']] /= mapping[['pos', 'neg']].sum()
            # For corner cases, defaulting to WOE = 0 (meaning no info). To avoid division by 0 we use default values.
            undef = (mapping['count'] < self.min_samples) | (mapping['pos'] == 0) | (mapping['neg'] == 0)
            mapping.loc[undef, ['pos', 'neg']] = -1
            # Final step, log of ratio of probabily estimates
            mapping['value'] = np.log(mapping['pos'] / mapping['neg'])
            self._mapping[col] = mapping



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import unicode_literals

import numpy as np

from mlencoders.base_encoder import BaseEncoder
from mlencoders.base_encoder import NAN_CATEGORY


class TargetEncoder(BaseEncoder):
    """
    Target Encoder for categorical features.
    """

    def __init__(self, cols=None, handle_unseen='impute', min_samples=1, smoothing=1):
        """Instantiation

        :param [str] cols: list of columns to encode, or None (then all dataset columns will be encoded at fitting time)
        :param str handle_unseen:
            'impute' - default value, impute a -1 category
            'error'  - raise an error if a category unseen at fitting time is found
            'ignore' - skip unseen categories
        :param int min_samples: minimum samples to take category average into account, must be >= 1
        :param int smoothing: coefficient used to balance categorical average (posterior) vs prior,
            the higher this number, the higher the prior is taken into account in the average

        :return: None
        """
        super(TargetEncoder, self).__init__(cols, handle_unseen, min_samples, None)
        self.smoothing = smoothing

    def fit(self, X, y):
        """Encode given columns of X according to y.

        :param pandas.DataFrame X: DataFrame of features, shape (n_samples, n_features). Must contain columns to encode.
        :param pandas.Series y: pandas Series of target values, shape (n_samples,).

        :return: None
        """
        self._before_fit_check(X, y)
        self._imputed = y.mean()
        for col in self.cols:
            mapping = y.groupby(X[col].fillna(NAN_CATEGORY)).agg(['mean', 'count'])
            corr_count = mapping['count'] - self.min_samples
            coef = (corr_count > 0) / (1 + np.exp(-corr_count / self.smoothing))
            mapping['value'] = self._imputed * (1 - coef) + mapping['mean'] * coef
            self._mapping[col] = mapping



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import unicode_literals

import cPickle as pickle

import numpy as np

NAN_CATEGORY = -99999


class BaseEncoder(object):

    def __init__(self, cols, handle_unseen, min_samples, imputed):
        self.cols = cols
        self.handle_unseen = handle_unseen
        self.min_samples = max(1, min_samples)
        # In case of unseen value or not enough data to learn the mapping, we use this value for imputation
        self._imputed = imputed
        # dict {str: pandas.DataFrame} column name --> mapping from category (index of df) to value (column of df)
        self._mapping = {}

    def transform(self, X):
        """Transform categorical data based on mapping learnt at fitting time.

        :param pandas.DataFrame X: DataFrame of features, shape (n_samples, n_features). Must contain columns to encode.

        :return: encoded DataFrame of shape (n_samples, n_features), initial categorical columns are dropped, and
            replaced with encoded columns. DataFrame passed in argument is unchanged.
        :rtype: pandas.DataFrame
        """
        if not self._mapping:
            raise ValueError('`fit` method must be called before `transform`.')
        assert all(c in X.columns for c in self.cols)

        X_encoded = X.copy(deep=True)
        for col, mapping in self._mapping.items():
            X_encoded[col] = mapping['value'].loc[X_encoded[col].fillna(NAN_CATEGORY)].values

            if self.handle_unseen == 'impute':
                X_encoded[col].fillna(self._imputed, inplace=True)
            elif self.handle_unseen == 'error':
                if np.unique(X_encoded[col]).shape[0] > mapping.shape[0]:
                    raise ValueError('Unseen categories found in `{}` column.'.format(col))

        return X_encoded

    def fit_transform(self, X, y):
        """Encode given columns of X according to y, and transform X based on the learnt mapping.

        :param pandas.DataFrame X: DataFrame of features, shape (n_samples, n_features). Must contain columns to encode.
        :param pandas.Series y: pandas Series of target values, shape (n_samples,).

        :return: encoded DataFrame of shape (n_samples, n_features), initial categorical columns are dropped, and
            replaced with encoded columns. DataFrame passed in argument is unchanged.
        :rtype: pandas.DataFrame
        """
        self.fit(X, y)
        return self.transform(X)

    def _before_fit_check(self, X, y):
        if self.cols is None:
            self.cols = X.columns
        else:
            assert all(c in X.columns for c in self.cols)
        assert X.shape[0] == y.shape[0]

    def save_as_object_file(self, path):
        if not self._mapping:
            raise ValueError('`fit` method must be called before `save_as_object_file`.')
        pickle.dump(self.__dict__, open(path, 'wb'))

    def load_from_object_file(self, path):
        for k, v in pickle.load(open(path, 'rb')).items():
            setattr(self, k, v)



# -*- coding: utf-8 -*-
from distutils.core import setup


version = '1.2.0'

setup(
    name='mlencoders',
    packages=['mlencoders'],
    version=version,
    description='''
        Machine Learning encoders for feature transformation & engineering: target encoder, weight of evidence.
    ''',
    url='https://github.com/tcassou/mlencoders',
    download_url='https://github.com/tcassou/mlencoders/archive/{}.tar.gz'.format(version),
    keywords=[
        'machine', 'learning', 'encoder', 'python',
        'feature', 'transform', 'engineering',
        'target', 'likelihood', 'weight', 'evidence',
        ],
    classifiers=[
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    install_requires=[
        'numpy>=1.14.0',
        'pandas>=0.22.0',
    ],
)
